# colour list
from typing import Any

c_colors = {
    "HEADER": "\033[95m",
    "OKBLUE": "\033[94m",
    "OKCYAN": "\033[96m",
    "OKGREEN": "\033[92m",
    "WARNING": "\033[93m",
    "FAIL": "\033[91m",
    "BOLD": "\033[1m",
    "UNDERLINE": "\033[4m",
}
end_c = "\033[0m"


def cstr(x: Any, ctype: str) -> str:
    return f"{c_colors[ctype]}{x}{end_c}"


log_flag = cstr(cstr("wingman", "BOLD"), "OKCYAN")


def wm_print(x: Any):
    print(f"{log_flag}: {x}")



import glob
import os
import shutil
import sys

from .print_utils import cstr, wm_print


def _get_dir_size(path):
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += _get_dir_size(entry.path)
    return total


def compress_weights():
    """Compresses the weights directory by deleting empty directories and compressing all mark numbers down to 0.

    Example CLI usage:
    `wingman-compress-weights [optional weights directory]`
    """
    wm_print("---------------------------------------------")
    wm_print("Beginning compression...")
    wm_print("---------------------------------------------")

    target_dir = "./weights"
    if len(sys.argv) > 1:
        target_dir = sys.argv[1]

    # get the original filesize
    original_size = _get_dir_size(target_dir)

    # all versions
    dirs = glob.glob(f"./{target_dir}/Version*")

    # which folders to delete
    to_delete = []

    # start going through each folder
    for dir in dirs:
        weights = glob.glob(os.path.join(dir, "weights*.pth"))

        # if no weights files
        if len(weights) == 0:
            to_delete.append(dir)
            continue

        # start finding the latest file
        num = 0
        has_num = True
        latest = ""
        while has_num:
            latest = os.path.join(dir, f"weights{num}.pth")

            # find the latest version and remove it from the list
            if latest in weights:
                num += 1
                has_num = True
            else:
                num -= 1
                has_num = False
                latest = os.path.join(dir, f"weights{num}.pth")
                weights.remove(latest)

        # remove the intermediary from our list if it's there
        intermediary = os.path.join(dir, "weights-1.pth")
        if intermediary in weights:
            weights.remove(intermediary)

        if len(weights) > 0:
            wm_print(f"Compressing {dir}...")

            # for the remainder of the list, delete all the weights
            for weight in weights:
                os.remove(weight)

            # rename the latest weights to 0
            os.rename(latest, os.path.join(dir, "weights0.pth"))

            wm_print(f"Compressed {cstr(dir, 'WARNING')}.")
        else:
            wm_print(f"Nothing to compress in {cstr(dir, 'OKGREEN')}.")

    # delete flagged directories
    for dir in to_delete:
        shutil.rmtree(dir, ignore_errors=False, onerror=None)
        wm_print(f"Deleted {cstr(dir, 'WARNING')}.")

    # get the final filesize
    final_size = _get_dir_size(target_dir)

    # printout
    wm_print("---------------------------------------------")
    wm_print(f"Original disk usage: {cstr(original_size / 1e9, 'OKGREEN')} gigabytes.")
    wm_print(
        f"Removed {cstr((original_size - final_size) / 1e9, 'WARNING')} gigabytes."
    )
    wm_print(f"Final disk usage: {cstr(final_size / 1e9, 'WARNING')} gigabytes.")
    wm_print("If this is unsatisactory, consider deleting weight directories manually.")
    wm_print("---------------------------------------------")


def generate_yaml():
    """Generates the basic yaml file for the Wingman module.

    Example CLI usage:
    `wingman-generate-yaml [optional filename]`
    """
    source = os.path.join(os.path.dirname(__file__), "./config.yaml")

    target = "./config.yaml"
    if len(sys.argv) > 1:
        target = sys.argv[1]

    shutil.copyfile(source, target)

    wm_print(f"Generated config file at: {cstr(target, 'OKGREEN')}.")



from .cli_scripts import compress_weights, generate_yaml
from .neural_blocks import NeuralBlocks
from .replay_buffer import ReplayBuffer
from .utils import cpuize, gpuize, shutdown_handler
from .wingman import Wingman



from __future__ import annotations

import numpy as np
from torch.utils.data import Dataset

from .print_utils import cstr, wm_print


class ReplayBuffer(Dataset):
    """Replay Buffer implementation of a Torch dataset."""

    def __init__(self, mem_size):
        self.memory = []
        self.mem_size = int(mem_size)
        self.count = 0

    def __len__(self):
        return min(self.mem_size, self.count)

    def __getitem__(self, idx):
        return (d[idx] for d in self.memory)

    def push(self, data: list[np.ndarray | float | int | bool], bulk: bool = False):
        # check if we are bulk adding things in and assert lengths
        bulk_size = 1
        if bulk:
            assert all([isinstance(data, np.ndarray)]), cstr(
                "All things must be np.ndarray for bulk data.", "FAIL"
            )

            bulk_size = data[0].shape[0]  # pyright: ignore

            assert all([len(d) == bulk_size for d in data]), cstr(  # pyright: ignore
                f"All things in data must have same len for the first dimension for bulk data. Received data with {[len(d) for d in data]} items respectively.",  # pyright: ignore
                "FAIL",
            )

        # expand dims of things that only have 1 dim
        def _ensure_dims(thing) -> np.ndarray:
            thing = np.asarray(thing)
            if len(thing.shape) == int(bulk):
                thing = np.expand_dims(thing, axis=-1)
            return thing

        np_data = list(map(_ensure_dims, data))

        # instantiate the memory if it does not exist
        if self.count == 0:
            self.memory = []
            for thing in np_data:
                if not bulk:
                    self.memory.append(
                        np.zeros((self.mem_size, *thing.shape), dtype=np.float32)
                    )
                else:
                    self.memory.append(
                        np.zeros((self.mem_size, *thing.shape[1:]), dtype=np.float32)
                    )

            mem_size = sum([d.nbytes for d in self.memory])
            wm_print(cstr(f"Replay Buffer Size: {mem_size / 1e9} gigabytes.", "OKCYAN"))

        # assert that the number of lists in memory is same as data to push
        assert len(np_data) == len(self.memory), cstr(
            f"Data length not similar to memory buffer length. Replay buffer has {len(self.memory)} items, but received {len(np_data)} items.",
            "FAIL",
        )

        # put stuff in memory
        i = self.count % self.mem_size
        for memory, thing in zip(self.memory, np_data):
            if not bulk:
                memory[i] = thing
            else:
                # get the remaining space of the replay buffer and
                # whether we need to wrap around
                remaining_space = self.mem_size - i
                front = min(remaining_space, bulk_size)
                back = remaining_space - bulk_size

                # add stuff to the buffer
                memory[i : i + front] = thing[:front]
                if back < 0:
                    memory[: abs(back)] = thing[back:]

        self.count += bulk_size

    @property
    def is_full(self) -> bool:
        return self.count >= self.mem_size



from setuptools import setup


def get_version():
    """Gets the pettingzoo version."""
    path = "pyproject.toml"
    with open(path) as file:
        lines = file.readlines()

    for line in lines:
        if line.startswith("version"):
            return line.strip().split()[-1].strip().strip('"')
    raise RuntimeError("bad version data in __init__.py")


setup(
    name="Wingman",
    version=get_version(),
    author="Jet",
    author_email="taijunjet@hotmail.com",
    description="Wingman for all your AI applications",
    url="https://github.com/jjshoots/Wingman",
    license_files=("LICENSE.txt"),
    long_description="Wingman library for AI projects.",
    long_description_content_type="text/markdown",
    keywords=["Machine Learning"],
    python_requires=">=3.7",
    packages=["wingman"],
    include_package_data=True,
    install_requires=["numpy", "torch", "wandb", "pyyaml"],
)
